/*
 * @DEC_COPYRIGHT@
 */
/*
 * HISTORY
 * $Log: socketvar.h,v $
 * Revision 4.2.85.4  2000/07/25  21:33:48  Tony_Bono
 * 	fixed sbflush panic when sb was locked by
 * 	creating a flag that can be set to indicate
 * 	that the data on a socket buf should be
 * 	flushed by the unlocker.
 *
 * Revision 4.2.85.3  2000/06/23  22:42:56  Tony_Bono
 * 	Modifications to eliminate the need to take the socket
 * 	lock when processing an accept for a tcp connection.
 *
 * Revision 4.2.85.2  2000/03/17  15:13:55  Brian_Haley
 * 	NUMA replication work - sb_max
 *
 * Revision 4.2.85.1  1999/12/07  23:21:29  Tony_Bono
 * 	Added per rad support to in_pcb logic
 *
 * Revision 4.2.34.11  1999/04/15  21:09:57  Tony_Bono
 * 	Implemented socket read lock support.
 * 	[1999/04/15  20:38:24  Tony_Bono]
 *
 * Revision 4.2.34.10  1999/03/30  23:47:47  Tony_Bono
 * 	Cache alignment changes.
 * 	[1999/03/30  23:38:16  Tony_Bono]
 * 
 * Revision 4.2.34.9  1999/01/20  19:08:02  Tony_Bono
 * 	 	** Merge Information **
 * 		** Command used:	bmerge -jPTEV6OS_BL8:PTFCOS_BL3 **
 * 		** Ancestor revision:	4.2.42.8 **
 * 		** Merge revision:	4.2.76.3 **
 * 	 	** End **
 * 	[1999/01/20  19:06:05  Tony_Bono]
 * 
 * Revision 4.2.76.3  1998/11/02  21:24:17  Tony_Bono
 * 	Increased SB_MAX to 1MB instead of 128K
 * 	[1998/10/30  16:22:50  Tony_Bono]
 * 
 * Revision 4.2.76.2  1998/09/29  22:25:22  Brian_Haley
 * 	Changed sb_timeo to be a struct timeval for Posix 1003.1g Sockets
 * 	support.
 * 	[1998/09/28  20:35:41  Brian_Haley]
 * 
 * Revision 4.2.34.8  1998/09/01  16:24:27  Vladislav_Yasevich
 * 	Changed the implementation to be unix98 complient. We are now storing the
 * 	timeval structure in the socket structure.  QAR 63422
 * 	[1998/08/24  17:58:58  Vladislav_Yasevich]
 * 
 * Revision 4.2.34.7  1998/01/21  21:15:20  Tony_Bono
 * 	 	** Merge Information **
 * 		** Command used:	bmerge -jPTCOS_BL11:PTCOS_BL14 **
 * 		** Ancestor revision:	4.2.42.7 **
 * 		** Merge revision:	4.2.42.8 **
 * 	 	** End **
 * 	[1998/01/21  20:53:54  Tony_Bono]
 * 
 * Revision 4.2.42.8  1997/10/09  23:22:35  Tony_Bono
 * 	Added the definitions of Flags and fields used to
 * 	support caching of checksums and wiring user pages
 * 	for output.
 * 	[1997/10/09  15:40:29  Tony_Bono]
 * 
 * Revision 4.2.34.6  1997/09/10  23:31:41  John_Dustin
 * 	 	** Merge Information **
 * 		** Command used:	bmerge -jPTCOS_BL6:PTCOS_BL10 **
 * 		** Ancestor revision:	4.2.42.4 **
 * 		** Merge revision:	4.2.42.6 **
 * 	 	** End **
 * 	merge of PTmin BL10 changes
 * 	[1997/09/10  23:04:50  John_Dustin]
 * 
 * Revision 4.2.42.7  1997/08/28  16:52:37  Tony_Bono
 * 	 	** Merge Information **
 * 		** Command used:	bmerge -jV40BSUPPORTOS_BL4:V40CSUPPORTOS_BL6 **
 * 		** Ancestor revision:	4.2.15.4 **
 * 		** Merge revision:	4.2.54.2 **
 * 	 	** End **
 * 	[1997/08/27  19:26:51  Tony_Bono]
 * 
 * Revision 4.2.42.6  1997/07/18  19:51:23  Kris_Corwin
 * 	Add SB_SET_LOWAT flag so system doesn't overwrite user settings.
 * 	[1997/07/09  15:26:16  Kris_Corwin]
 * 
 * Revision 4.2.42.5  1997/05/30  22:35:50  Tony_Bono
 * 	HPXQB6189/QAR 49847, An enhanced fix to solockpair() routine
 * 	was needed because the routine was freeing a solock structure
 * 	that was concurrently spun upon in lock_write().
 * 	[1997/05/30  22:17:31  Tony_Bono]
 * 
 * Revision 4.2.54.2  1997/04/25  19:29:55  Benjamin_Dibble
 * 	Changes for Mondo2.
 * 	inline imin() call for sbspace so that soo_select() routine
 * 	will become a leaf routine.
 * 	Changes to make socket select a readonly operation...
 * 	Hint code support.
 * 	[1997/04/24  19:27:47  Benjamin_Dibble]
 * 
 * Revision 4.2.34.5  1997/04/24  18:27:23  Tony_Bono
 * 	 	** Merge Information **
 * 		** Command used:	bmerge -jPTCOS_BL3:PTCOS_BL6 **
 * 		** Ancestor revision:	4.2.42.2 **
 * 		** Merge revision:	4.2.42.4 **
 * 	 	** End **
 * 	merge
 * 	[1997/04/24  18:09:20  Tony_Bono]
 * 
 * Revision 4.2.42.4  1997/03/06  17:10:29  James_Woodward
 * 	inline imin() call for sbspace so that soo_select() routine
 * 	will become a leaf routine.
 * 	[1997/03/05  12:18:37  James_Woodward]
 * 
 * Revision 4.2.34.4  1997/02/06  02:47:13  Tony_Bono
 * 	 	** Merge Information **
 * 		** Command used:	bmerge -jPTBOS_BL4:PTCOS_BL3 **
 * 		** Ancestor revision:	4.2.15.4 **
 * 		** Merge revision:	4.2.42.2 **
 * 	 	** End **
 * 	merge of PTC BL3 changes
 * 	[1997/02/06  02:31:54  Tony_Bono]
 * 
 * Revision 4.2.42.3  1997/01/23  18:33:59  James_Woodward
 * 	changes to make socket select a readonly operation...
 * 	[1997/01/15  12:45:09  James_Woodward]
 * 
 * Revision 4.2.42.2  1996/11/22  22:11:23  Tony_Bono
 * 	hint code support
 * 	[1996/11/22  22:02:45  Tony_Bono]
 * 
 * Revision 4.2.34.3  1996/11/19  21:47:57  Tony_Bono
 * 	final pass of web server package including hint,
 * 	 quick bind and tcp syn queue changes
 * 	[1996/11/14  16:44:36  Tony_Bono]
 * 
 * Revision 4.2.36.2  1996/10/10  18:15:01  Tony_Bono
 * 	hint patch initial rev 10-10-96
 * 
 * Revision 4.2.34.2  1996/09/25  18:48:25  Brian_Haley
 * 	Added C++ forward declaration
 * 	[1996/09/24  20:09:52  Brian_Haley]
 * 
 * Revision 4.2.15.4  1996/02/29  15:06:08  Ajay_Kachrani
 * 	Add SOCKET_TRY_WRITE_LOCK() macro.
 * 	[1996/02/20  19:52:39  Ajay_Kachrani]
 * 
 * Revision 4.2.15.3  1995/04/14  21:23:10  John_Dustin
 * 	DART hooks: added sb_dbcnt to sbspace(), sballoc() and sbfree();
 * 	added SB_DBMAX_DEF flag
 * 	[1995/04/14  19:06:03  John_Dustin]
 * 
 * Revision 4.2.15.2  1995/03/24  21:21:51  Jack_McCann
 * 	Add SP_XSE.
 * 	[1995/03/22  22:07:06  Jack_McCann]
 * 
 * Revision 4.2.6.9  1994/04/06  19:09:55  Ernie_Petrides
 * 	Integrate solock() and sounlock() with the call patch-out mechanism,
 * 	change position of so_timeo to allow so_error to become an int, remove
 * 	so_ptcb field, clean up NETSYNC_LOCKTEST, and update comments for
 * 	obsolete SP_PIPE and SP_WATOMIC.
 * 	[1994/04/06  12:48:05  Ernie_Petrides]
 * 
 * Revision 4.2.6.8  1994/03/18  21:54:26  Ernie_Petrides
 * 	Convert locking calls for socklocks.*_lock to use the ulock_*
 * 	versions to allow for call patch-out in lockmode 0.
 * 	[1994/03/17  00:30:43  Ernie_Petrides]
 * 
 * Revision 4.2.6.7  1994/03/02  00:01:17  Geraldine_Harter
 * 	Reduce size of socket fields (long to int).
 * 	[1994/02/28  19:03:14  Geraldine_Harter]
 * 
 * Revision 4.2.6.6  1993/11/11  18:21:44  Geraldine_Harter
 * 	Fix SOCKET_ISLOCKED() & SOCKBUF_ISLOCKED() macros - test that lock
 * 	is held by caller (instead of testing only that the lock is held).
 * 	[1993/11/03  20:52:43  Geraldine_Harter]
 * 
 * Revision 4.2.6.5  1993/10/14  20:53:15  Ernie_Petrides
 * 	Add SOCKET_LOCKTERMINATE macro to fix socket lock instance counts.
 * 	[1993/10/14  20:41:53  Ernie_Petrides]
 * 
 * Revision 4.2.6.4  1993/10/07  21:59:53  Geraldine_Harter
 * 	Merge Sterling BL6:
 * 
 * 	Revision 4.2.9.2  1993/07/26  21:03:46  Heather_Gray
 * 	Change socket fields (so_type, so_options, so_state, so_special,
 * 	sb_flags and sb_timeo) from short to int for performance.
 * 	Removed so_spare.  Increased SB_MAX from 64K to 128K.
 * 	[1993/07/26  20:57:12  Heather_Gray]
 * 
 * 	Note: the following revision info was dropped in Gold.  Add it now
 * 	for completeness:
 * 
 * 	Revision 4.2.2.2  92/03/23  11:07:25  Benjamin_England
 * 	added support in soreadable for protocols which support notion of
 * 	zero-length reads.  Such protocols transmit a series of records,
 * 	rather than a series of bytes.
 * 	[92/03/10  20:47:13  Benjamin_England]
 * 
 * Revision 4.2.6.3  1993/08/03  16:36:08  Ajay_Kachrani
 * 	Convert n/w locks to new lock stats using LI
 * 	[1993/07/22  19:38:23  Ajay_Kachrani]
 * 
 * Revision 4.2.6.2  1993/06/25  21:20:27  Geraldine_Harter
 * 	Merge OSF1.2 changes.
 * 	[1993/06/25  20:37:11  Geraldine_Harter]
 * 
 * Revision 4.2.4.2  1992/04/14  13:43:41  Al_Delorey
 * 	From Silver: AG_BL6
 * 	[92/04/14  13:42:58  Al_Delorey]
 * 
 * Revision 4.2.2.2  92/04/13  17:56:33  devbld_gsf
 * 	First AG_BL6 version.  Moved from Ag baselevel delivery tree.
 * 
 * Revision 4.2  91/09/19  23:00:19  devbld
 * 	Adding ODE Headers
 * 
 * $EndLog$
 */
/*	
 *	@(#)$RCSfile: socketvar.h,v $ $Revision: 4.2.85.4_51_BL17_TERTENTELLA_CLIFF $ (DEC) $Date: 2001/05/07 10:00:00 $
 */ 
/*
 */
/*
 * (c) Copyright 1990, OPEN SOFTWARE FOUNDATION, INC.
 * ALL RIGHTS RESERVED
 */
/*
 * OSF/1 Release 1.0
 */

/*
 * Copyright (C) 1988,1989 Encore Computer Corporation.  All Rights Reserved
 *
 * Property of Encore Computer Corporation.
 * This software is made available solely pursuant to the terms of
 * a software license agreement which governs its use. Unauthorized
 * duplication, distribution or sale are strictly prohibited.
 *
 */
/*
 * Copyright (c) 1982, 1986, 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted provided
 * that: (1) source distributions retain this entire copyright notice and
 * comment, and (2) distributions including binaries display the following
 * acknowledgement:  ``This product includes software developed by the
 * University of California, Berkeley and its contributors'' in the
 * documentation or other materials provided with the distribution and in
 * all advertising materials mentioning features or use of this software.
 * Neither the name of the University nor the names of its contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	Base:	socketvar.h	7.6 (Berkeley) 9/4/89
 *	Merged: socketvar.h	7.11 (Berkeley) 6/30/90
 */

#ifndef	_SYS_SOCKETVAR_H_
#define	_SYS_SOCKETVAR_H_

#ifdef __cplusplus
extern "C" {
/* forward decls for C++ */
struct socklocks;
#endif

#if	defined(_KERNEL) && !defined(_NET_GLOBALS_H_)
#include "net/net_globals.h"
#define sb_lock_t	lock_t
#else
typedef struct lock *	sb_lock_t;
#endif

#include <sys/secdefines.h>
#if	SEC_ARCH
#include <sys/security.h>
#endif
#include <sys/time.h>

/*
 * Kernel structure per socket.
 * Contains send and receive buffer queues,
 * handle on protocol and pointer to protocol
 * private data and error information.
 */
struct socket {
	int	so_type;		/* generic type, see socket.h */
	int	so_options;		/* from socket call, see socket.h */
	short	so_linger;		/* time to linger while closing */
	short	so_timeo;		/* connection timeout */
	int	so_state;		/* internal state flags SS_*, below */
	caddr_t	so_pcb;			/* protocol control block */
	struct	protosw *so_proto;	/* protocol handle */
	struct	socklocks *so_lock;	/* socket structure lock(s) */
	int	so_atomic_flags;	/* atomic flags set and cleared using atomic ops */
	int pad_a[5];
/*
 * Variables for connection queueing.
 * Socket where accepts occur is so_head in all subsidiary sockets.
 * If so_head is 0, socket is not related to an accept.
 * For head socket so_q0 queues partially completed connections,
 * while so_q is a queue of connections ready to be accepted.
 * If a connection is aborted and it has so_head set, then
 * it has to be pulled out of either so_q0 or so_q.
 * We allow connections to queue up based on current queue lengths
 * and limit on number of queued connections for this socket.
 */
/* 64-byte alignment */
	struct	socket *so_head;	/* back pointer to accept socket */
	struct	socket *so_q0;		/* queue of partial connections */
	struct	socket *so_q;		/* queue of incoming connections */
	struct	socket *so_dq;		/* queue of defunct connections */
	unsigned short	so_q0len;		/* partials on so_q0 */
	unsigned short	so_qlen;		/* number of connections on so_q */
	unsigned short	so_qlimit;		/* max number queued connections */
	short	so_dqlen;		/* listener dequeues in progress */
	int	so_error;		/* error affecting connection */
	int	so_special;		/* special state flags SP_*, below */
	pid_t	so_pgid;		/* pgid for signals */
	u_int	so_oobmark;		/* chars to oob mark */
	int pad_b[1];
/* 64-byte alignment */
	struct	sockbuf {
		u_int	sb_cc;		/* actual chars in buffer */
		u_int	sb_hiwat;	/* max actual char count */
		u_int	sb_mbcnt;	/* chars of mbufs used */
		u_int	sb_mbmax;	/* max chars of mbufs to use */
		u_int	sb_dbcnt;	/* DART: chars of dart-style used */
		u_int	sb_dbmax;	/* DART: max chars to use */
		int	sb_lowat;	/* low water mark */
		int     sb_sel;         /* select hint */	
		struct	mbuf *sb_mb;	/* the mbuf chain */
		struct sbselque {	/* process selecting read/write */
			struct sbselque *next, *prev;
		} sb_selq;
		int	sb_flags;	/* flags, see below */
		int pad_c[1];

/* 64-byte alignment */
		struct timeval	sb_timeo;	/* timeout value in seconds and microsec */
#if	__STDC__
		void	(*sb_wakeup)(caddr_t, int);
#else
		void	(*sb_wakeup)();	/* upcall instead of sowakeup */
#endif
		caddr_t	sb_wakearg;	/* (*sb_wakeup)(sb_wakearg, state) */
		sb_lock_t sb_lock;	/* sockbuf lock (in socklocks) */
		struct mbuf *sb_last_mb;/* last mbuf on the current chain only valid if sb_mb set */
		int pad_d[6];
	} so_rcv, so_snd;
#define	SB_MAX		(1024*1024)	/* default for max chars in sockbuf */
#define	SB_LOCK		0x01		/* lock on data queue */
#define	SB_WANT		0x02		/* someone is waiting to lock */
#define	SB_WAIT		0x04		/* someone is waiting for data/space */
#define	SB_SEL		0x08		/* someone is selecting */
#define	SB_ASYNC	0x10		/* ASYNC I/O, need signals */
#define	SB_NOTIFY	(SB_WAIT|SB_SEL|SB_ASYNC)
#define	SB_COLL		0x20		/* collision selecting (UNIX) */
#define	SB_NOINTR	0x40		/* operations not interruptible */
#define	SB_WAKEONE	0x80		/* wakeup only one on notify */
#define	SB_WAITING	0x100		/* MSG_WAITALL receive in progress */
#define	SB_SET_LOWAT	0x200		/* user set sb_lowat */
#define	SB_WIRE		0x400		/* Wire don't copy user pages(currently output only) */
#define SB_NEEDSFLUSH	0x800		/* next unlock flush buffer */
#define SB_USELAST  0x1000      /* code expects the use of sb_last_m */
#define	SB_INHERIT	(SB_NOINTR|SB_WAKEONE|SB_WIRE)
#define	SB_DBMAX_DEF	(1 << 30)	/* DART */
/* 64-byte alignment */
#if	SEC_ARCH
	tag_t	so_tag[SEC_TAG_COUNT];	/* security policy tags */
#endif
	struct	socket *so_q0_prev;	/* end of queue of partial connections */
	struct	socket *so_q_prev;	/* end of queue of incoming connections */
	int pad_e[4];
};

/*
 * Socket state bits.
 */
#define	SS_NOFDREF		0x001	/* no file table ref any more */
#define	SS_ISCONNECTED		0x002	/* socket connected to a peer */
#define	SS_ISCONNECTING		0x004	/* in process of connecting to peer */
#define	SS_ISDISCONNECTING	0x008	/* in process of disconnecting */
#define	SS_CANTSENDMORE		0x010	/* can't send more data to peer */
#define	SS_CANTRCVMORE		0x020	/* can't receive more data from peer */
#define	SS_RCVATMARK		0x040	/* at mark on input */

#define	SS_PRIV			0x080	/* privileged for broadcast, raw... */
#define	SS_NBIO			0x100	/* non-blocking ops */
#define	SS_ASYNC		0x200	/* async i/o notify */
#define	SS_ISCONFIRMING		0x400	/* deciding to accept connection req */

/*
 * Socket flags set using atomic ops to set and clear bits
 */
#define SOF_FDREF		0x001	/* atomic flag to indicate File ref set for this connection */
#define SOF_FDREF_BIT		0x000	/* atomic flag bit to indicate File ref set for this connection */
#define SOF_CLEARING		0x002	/* atomic flag to indicate Listener is closing and clearing accepted connections */
#define SOF_CLEARING_BIT	0x001	/* atomic flag bit to indicate Listener is closing and clearing accepted connections */
#define SOF_PRIV		0x004	/* atomic flag to indicate user accout is privilaged */
#define SOF_PRIV_BIT		0x002	/* atomic flag bit to indicate user accout is privilaged */
#define SOF_PROTO_USE		0x800	/* atomic flag for use by protocol */
#define SOF_PROTO_USE_BIT	0x800	/* atomic flag bit for use by protocol */

/*
 * Special socket state bits.
 */
#define SP_PIPE			0x0001	/* socket is unnamed pipe (obsolete) */
#define SP_WATOMIC		0x0002	/* pipe write atomicity (obsolete) */
#define SP_NOUAREA		0x0004	/* no u-area available (XTI - XXX) */
#define SP_LOCKABLE		0x0008	/* locking is active for socket */
#define SP_CLOSING		0x0010	/* closing a listening socket */
#define SP_RWAKEUP		0x0020	/* sorwakeup pending release of lock */
#define SP_WWAKEUP		0x0040	/* sowwakeup pending release of lock */
#define SP_EXTPRIV		0x0080	/* SS_PRIV managed externally */
#define SP_XSE			0x0100	/* _XOPEN_SOURCE_EXTENDED socket */
#define SP_NBIO			0x0200  /* force nbio for any children */
#define SP_FREEABLE		0x8000	/* free socket on unlock */

#define SP_INHERIT	(SP_PIPE|SP_WATOMIC|SP_NOUAREA|SP_LOCKABLE|SP_XSE|SP_NBIO)

#ifdef	_KERNEL

#ifndef REPLICATED
#define REPLICATED
#endif

/*
 * Macros for sockets, socket locking and socket buffering.
 */

#if	NETSYNC_LOCK

struct  socklocks {
	lock_data_t	sock_lock;
	lock_data_t	snd_lock;
	lock_data_t	rcv_lock;
	int		refcnt;
	struct socket	*sp_wake;
	struct socket	*socketlist ; 
	struct socklocks *nextlock ;
	int pad[2];
};

#define SOCKET_LOCKINIT(so, lp) { \
	(so)->so_lock = (lp); \
	(so)->so_snd.sb_lock = &((lp)->snd_lock); \
	(so)->so_rcv.sb_lock = &((lp)->rcv_lock); \
	(lp)->nextlock = NULL ; \
	(lp)->socketlist = so ; \
	(lp)->refcnt = 1; \
	(lp)->sp_wake = 0; \
	ulock_setup(&((lp)->sock_lock), sock_li, TRUE); \
	ulock_setup(&((lp)->snd_lock), snd_li, TRUE); \
	ulock_setup(&((lp)->rcv_lock), rcv_li, TRUE); \
}

#define SOCKET_LOCKTERMINATE(lp) { \
	ulock_terminate(&(lp)->sock_lock); \
	ulock_terminate(&(lp)->snd_lock); \
	ulock_terminate(&(lp)->rcv_lock); \
}

#define SOCKLOCK_REF(lp) {  \
	simple_lock(&lp->sock_lock.l_lock) ;  \
        lp->refcnt++ ; \
	simple_unlock(&lp->sock_lock.l_lock) ;  \
}

/*
 * In lockmode 0, calls to solock() and sounlock() are dynamically
 * patched out.  
 */
#define SOCKET_LOCK(so)		solock(so)	/* noop lockmode 0 */
#define SOCKET_UNLOCK(so)	sounlock(so)	/* noop lockmode 0 */

#define SOCKET_READ_LOCK(so)	soreadlock(so)	/* noop lockmode 0 */
#define SOCKET_READ_UNLOCK(so)	soreadunlock(so)	/* noop lockmode 0 */

#define SOCKET_LOCK_READERS(so)	soreaders(so)	/* noop lockmode 0 */

#define SOCKET_TRY_WRITE_LOCK(so) solocktry(so) /* noop lockmode 0 */
#define SOCKET_TRY_WRITE_NOASSERT(so) solocktrynoassert(so) /* noop lockmode 0 */

#define SOCKET_LOCK_ONE(so1) solock1(so1)

#define SOCKET_UNLOCK_ONE(so1) sounlock1(so1)

/*
 * Lock order of unpaired sockets is critical to avoid deadlock.
 */
#define SOCKET_LOCK2(so1, so2) { \
	SOCKET_LOCK_ONE(so1) ; \
	if((!SOCKET_ISLOCKED((so2))) && (!SOCKET_TRY_WRITE_NOASSERT(so2))) {\
		if(so1->so_lock > so2->so_lock)  {\
		      SOCKET_LOCK_ONE(so2); \
		} else {\
			SOCKET_UNLOCK_ONE(so1) ; \
			SOCKET_LOCK_ONE(so2) ; \
			SOCKET_LOCK_ONE(so1) ; \
		} \
	} \
}

#define SOCKET_UNLOCK2(so1, so2) {  \
        sounlock((so1));         \
	if ( SOCKET_ISLOCKED(so2))  \
	       SOCKET_UNLOCK_ONE(so2);  \
}

#define SOCKET_UNLOCK_HALF_OF_2(so1, so2) { \
        if( (so1)->so_lock != (so2)->so_lock ) /* same lock */ \
		   SOCKET_UNLOCK_ONE(so2); \
}
/*
 * Lock order of unpaired sockets is critical to avoid deadlock.
 */
#define SOCKET_LOCK_HALF_OF_2(so1, so2) { \
	if ( (!SOCKET_ISLOCKED((so2))) && (!SOCKET_TRY_WRITE_NOASSERT(so2)) ) { \
		/* quick didn't work now have to maintain order */ \
		if(so1->so_lock > so2->so_lock)  { \
		      SOCKET_LOCK_ONE(so2); \
		} else {\
			SOCKET_UNLOCK_ONE(so1) ; \
			SOCKET_LOCK_ONE(so2) ; \
			SOCKET_LOCK_ONE(so1) ; \
		} \
	} \
}

/* The socket is considered locked for assertions if not lockable */
#define SOCKET_ISLOCKED(so)  (!(so->so_special & SP_LOCKABLE) || \
				lock_holder(&((so)->so_lock->sock_lock)))

/* sockbuf locks are currently disabled */
#define SOCKBUF_LOCKTEST	0
#define SOCKBUF_LOCK(sb)	/*ulock_write((sb)->sb_lock)*/
#define SOCKBUF_UNLOCK(sb)	/*ulock_done((sb)->sb_lock)*/
#define SOCKBUF_ISLOCKED(sb)	1 /*(lockmode==0||lock_holder((sb)->sb_lock))*/

#else	/* !NETSYNC_LOCK */

struct  socklocks { int refcnt; };
#define SOCKET_LOCKINIT(so, lp)
#define SOCKET_LOCKTERMINATE(lp)
#define SOCKET_UNLOCK_HALF_OF_2(so1, so2)
#define SOCKET_LOCK_HALF_OF_2(so1, so2)
#define SOCKET_READ_LOCK(so)
#define SOCKET_LOCK(so)
#define SOCKET_LOCK_ONE(so)
#define SOCKET_TRY_WRITE_LOCK(so)	1
#define SOCKET_TRY_WRITE_NOASSERT(so)  1
#define SOCKET_READ_UNLOCK(so)
#define SOCKET_LOCK_READERS(so)	 1
#define SOCKET_UNLOCK(so)
#define SOCKET_UNLOCK_ONE(so)
#define SOCKET_LOCK2(so1, so2)
#define SOCKET_UNLOCK2(so1, so2)
#define SOCKET_ISLOCKED(so)	1
#define SOCKBUF_LOCKTEST	0
#define SOCKBUF_LOCK(sb)
#define SOCKBUF_UNLOCK(sb)
#define SOCKBUF_ISLOCKED(sb)	1

#endif	/* !NETSYNC_LOCK */

/* Bits for network events to sb_wakeup */
#define SE_ERROR	0x0001	/* so_error non-0 */
#define SE_HAVEDATA	0x0002	/* data in send or recv q */
#define SE_HAVEOOB	0x0004	/* oob data in recv q */
#define SE_DATAFULL	0x0008	/* send or recv q is full */
#define SE_CONNOUT	0x0010	/* outgoing connect complete (connect) */
#define SE_CONNIN	0x0020	/* incoming connect complete (listen)  */
#define SE_SENDCONN	0x0040	/* connected for send */
#define SE_RECVCONN	0x0080	/* connected for recv */
#define SE_POLL		0x4000	/* wakeup is synchronous poll */
#define SE_STATUS	0x8000	/* above status bits valid */

/* does socket have a valid u-area associated? (for stats, security, etc) */
#define SOHASUAREA(so)	(!((so)->so_special & SP_NOUAREA))

/* what is the domain associated with this socket? */
#define sodomain(so)	((so)->so_proto->pr_domain)

/*
 * How much space is there in a socket buffer (so->so_snd or so->so_rcv)?
 * This is problematical if the fields are unsigned, as the space might
 * still be negative (cc > hiwat or mbcnt > mbmax).  Should detect
 * overflow and return 0.  Should use "lmin" but it doesn't exist now.
 */
#define __sbspace_imin(a,b)  ((a) < (b) ? (a) : (b))

#define sbspace(sb) \
    ((int) __sbspace_imin((int)((sb)->sb_hiwat - (sb)->sb_cc), \
	__sbspace_imin((int)((sb)->sb_mbmax - (sb)->sb_mbcnt), \
	(int)((sb)->sb_dbmax - (sb)->sb_dbcnt))))

/* do we have to send all at once on a socket? */
#define	sosendallatonce(so) \
    ((so)->so_proto->pr_flags & PR_ATOMIC)

/* Is write of len bytes to be atomic (for pipes)?  (obsolete) */
#define sowatomic(so, len) (0)

/* can we read something from so? 
 * for certain sequenced packet protocols, zero-length reads are supported 
 */
#define	soreadable(so) \
    ((so)->so_rcv.sb_cc >= (so)->so_rcv.sb_lowat || \
	((so)->so_state & SS_CANTRCVMORE) || \
	(so)->so_qlen || (so)->so_error) || \
	(((so)->so_proto->pr_flags & PR_READZEROLEN ) && \
	 ((so)->so_rcv.sb_mb))

/* can we write something to so? */
#define	sowriteable(so) \
    (sbspace(&(so)->so_snd) >= (so)->so_snd.sb_lowat && \
	(((so)->so_state&SS_ISCONNECTED) || \
	  ((so)->so_proto->pr_flags&PR_CONNREQUIRED)==0) || \
     ((so)->so_state & SS_CANTSENDMORE) || \
     (so)->so_error)

/* adjust counters in sb reflecting allocation of m */
#define sballoc(sb, m) { \
	(sb)->sb_cc += (m)->m_len; \
	(sb)->sb_mbcnt += MSIZE; \
	if ((m)->m_flags & M_EXT) { \
		if ((m)->m_flags & M_UIOMOVE) \
			(sb)->sb_dbcnt += (m)->m_ext.ext_size; \
		else \
			(sb)->sb_mbcnt += (m)->m_ext.ext_size; \
	} \
}

/* adjust counters in sb reflecting freeing of m */
#define sbfree(sb, m) { \
	(sb)->sb_cc -= (m)->m_len; \
	(sb)->sb_mbcnt -= MSIZE; \
	if ((m)->m_flags & M_EXT) { \
		if ((m)->m_flags & M_UIOMOVE) \
			(sb)->sb_dbcnt -= (m)->m_ext.ext_size; \
		else \
			(sb)->sb_mbcnt -= (m)->m_ext.ext_size; \
	} \
}

#define	sorwakeup(so)	sowakeup((so), &(so)->so_rcv)
#define	sowwakeup(so)	sowakeup((so), &(so)->so_snd)

/* to catch callers missing new second argument to sonewconn: */
#define sonewconn(head, connstatus)	sonewsock((head), (connstatus))
extern REPLICATED u_int	sb_max;
#endif	/* _KERNEL */

#ifdef __cplusplus
}
#endif
#endif
